# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: spec/prediction_job.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19spec/prediction_job.proto\x12\x11merlin.batch.spec\"\xd3\x03\n\rPredictionJob\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x0c\n\x04kind\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\x12<\n\x0f\x62igquery_source\x18\x0b \x01(\x0b\x32!.merlin.batch.spec.BigQuerySourceH\x00\x12\x32\n\ngcs_source\x18\x0c \x01(\x0b\x32\x1c.merlin.batch.spec.GcsSourceH\x00\x12@\n\x11maxcompute_source\x18\r \x01(\x0b\x32#.merlin.batch.spec.MaxComputeSourceH\x00\x12\'\n\x05model\x18\x15 \x01(\x0b\x32\x18.merlin.batch.spec.Model\x12\x38\n\rbigquery_sink\x18\x1f \x01(\x0b\x32\x1f.merlin.batch.spec.BigQuerySinkH\x01\x12.\n\x08gcs_sink\x18  \x01(\x0b\x32\x1a.merlin.batch.spec.GcsSinkH\x01\x12<\n\x0fmaxcompute_sink\x18! \x01(\x0b\x32!.merlin.batch.spec.MaxComputeSinkH\x01\x42\x08\n\x06sourceB\x06\n\x04sink\"\xa2\x01\n\x0e\x42igQuerySource\x12\r\n\x05table\x18\x01 \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\x02 \x03(\t\x12?\n\x07options\x18\x03 \x03(\x0b\x32..merlin.batch.spec.BigQuerySource.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xc5\x01\n\tGcsSource\x12-\n\x06\x66ormat\x18\x01 \x01(\x0e\x32\x1d.merlin.batch.spec.FileFormat\x12\x0b\n\x03uri\x18\x02 \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\x03 \x03(\t\x12:\n\x07options\x18\x04 \x03(\x0b\x32).merlin.batch.spec.GcsSource.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xb8\x01\n\x10MaxComputeSource\x12\r\n\x05table\x18\x01 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x02 \x01(\t\x12\x10\n\x08\x66\x65\x61tures\x18\x03 \x03(\t\x12\x41\n\x07options\x18\x04 \x03(\x0b\x32\x30.merlin.batch.spec.MaxComputeSource.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xcc\x02\n\x05Model\x12*\n\x04type\x18\x01 \x01(\x0e\x32\x1c.merlin.batch.spec.ModelType\x12\x0b\n\x03uri\x18\x02 \x01(\t\x12\x34\n\x06result\x18\x03 \x01(\x0b\x32$.merlin.batch.spec.Model.ModelResult\x12\x36\n\x07options\x18\x04 \x03(\x0b\x32%.merlin.batch.spec.Model.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1al\n\x0bModelResult\x12+\n\x04type\x18\x01 \x01(\x0e\x32\x1d.merlin.batch.spec.ResultType\x12\x30\n\titem_type\x18\x02 \x01(\x0e\x32\x1d.merlin.batch.spec.ResultType\"\xeb\x01\n\x0c\x42igQuerySink\x12\r\n\x05table\x18\x01 \x01(\t\x12\x16\n\x0estaging_bucket\x18\x02 \x01(\t\x12\x15\n\rresult_column\x18\x03 \x01(\t\x12.\n\tsave_mode\x18\x04 \x01(\x0e\x32\x1b.merlin.batch.spec.SaveMode\x12=\n\x07options\x18\x05 \x03(\x0b\x32,.merlin.batch.spec.BigQuerySink.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xf6\x01\n\x07GcsSink\x12-\n\x06\x66ormat\x18\x01 \x01(\x0e\x32\x1d.merlin.batch.spec.FileFormat\x12\x0b\n\x03uri\x18\x02 \x01(\t\x12\x15\n\rresult_column\x18\x03 \x01(\t\x12.\n\tsave_mode\x18\x04 \x01(\x0e\x32\x1b.merlin.batch.spec.SaveMode\x12\x38\n\x07options\x18\x05 \x03(\x0b\x32\'.merlin.batch.spec.GcsSink.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xe9\x01\n\x0eMaxComputeSink\x12\r\n\x05table\x18\x01 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x02 \x01(\t\x12\x15\n\rresult_column\x18\x03 \x01(\t\x12.\n\tsave_mode\x18\x04 \x01(\x0e\x32\x1b.merlin.batch.spec.SaveMode\x12?\n\x07options\x18\x05 \x03(\x0b\x32..merlin.batch.spec.MaxComputeSink.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*Q\n\nResultType\x12\n\n\x06\x44OUBLE\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\x0b\n\x07INTEGER\x10\x02\x12\x08\n\x04LONG\x10\x03\x12\n\n\x06STRING\x10\x04\x12\t\n\x05\x41RRAY\x10\n*\x7f\n\tModelType\x12\x16\n\x12INVALID_MODEL_TYPE\x10\x00\x12\x0b\n\x07XGBOOST\x10\x01\x12\x0e\n\nTENSORFLOW\x10\x02\x12\x0b\n\x07SKLEARN\x10\x03\x12\x0b\n\x07PYTORCH\x10\x04\x12\x08\n\x04ONNX\x10\x05\x12\n\n\x06PYFUNC\x10\x06\x12\r\n\tPYFUNC_V2\x10\x07*O\n\nFileFormat\x12\x17\n\x13INVALID_FILE_FORMAT\x10\x00\x12\x07\n\x03\x43SV\x10\x01\x12\x0b\n\x07PARQUET\x10\x02\x12\x08\n\x04\x41VRO\x10\x03\x12\x08\n\x04JSON\x10\x04*O\n\x08SaveMode\x12\x11\n\rERRORIFEXISTS\x10\x00\x12\r\n\tOVERWRITE\x10\x01\x12\n\n\x06\x41PPEND\x10\x02\x12\n\n\x06IGNORE\x10\x03\x12\t\n\x05\x45RROR\x10\x04\x42\x66\n\x1b\x63om.gojek.merlin.batch.specB\x12PredictionJobProtoP\x01Z1github.com/caraml-dev/merlin-pyspark-app/pkg/specb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'spec.prediction_job_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.gojek.merlin.batch.specB\022PredictionJobProtoP\001Z1github.com/caraml-dev/merlin-pyspark-app/pkg/spec'
  _globals['_BIGQUERYSOURCE_OPTIONSENTRY']._options = None
  _globals['_BIGQUERYSOURCE_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_GCSSOURCE_OPTIONSENTRY']._options = None
  _globals['_GCSSOURCE_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_MAXCOMPUTESOURCE_OPTIONSENTRY']._options = None
  _globals['_MAXCOMPUTESOURCE_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_MODEL_OPTIONSENTRY']._options = None
  _globals['_MODEL_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_BIGQUERYSINK_OPTIONSENTRY']._options = None
  _globals['_BIGQUERYSINK_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_GCSSINK_OPTIONSENTRY']._options = None
  _globals['_GCSSINK_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_MAXCOMPUTESINK_OPTIONSENTRY']._options = None
  _globals['_MAXCOMPUTESINK_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_RESULTTYPE']._serialized_start=2128
  _globals['_RESULTTYPE']._serialized_end=2209
  _globals['_MODELTYPE']._serialized_start=2211
  _globals['_MODELTYPE']._serialized_end=2338
  _globals['_FILEFORMAT']._serialized_start=2340
  _globals['_FILEFORMAT']._serialized_end=2419
  _globals['_SAVEMODE']._serialized_start=2421
  _globals['_SAVEMODE']._serialized_end=2500
  _globals['_PREDICTIONJOB']._serialized_start=49
  _globals['_PREDICTIONJOB']._serialized_end=516
  _globals['_BIGQUERYSOURCE']._serialized_start=519
  _globals['_BIGQUERYSOURCE']._serialized_end=681
  _globals['_BIGQUERYSOURCE_OPTIONSENTRY']._serialized_start=635
  _globals['_BIGQUERYSOURCE_OPTIONSENTRY']._serialized_end=681
  _globals['_GCSSOURCE']._serialized_start=684
  _globals['_GCSSOURCE']._serialized_end=881
  _globals['_GCSSOURCE_OPTIONSENTRY']._serialized_start=635
  _globals['_GCSSOURCE_OPTIONSENTRY']._serialized_end=681
  _globals['_MAXCOMPUTESOURCE']._serialized_start=884
  _globals['_MAXCOMPUTESOURCE']._serialized_end=1068
  _globals['_MAXCOMPUTESOURCE_OPTIONSENTRY']._serialized_start=635
  _globals['_MAXCOMPUTESOURCE_OPTIONSENTRY']._serialized_end=681
  _globals['_MODEL']._serialized_start=1071
  _globals['_MODEL']._serialized_end=1403
  _globals['_MODEL_OPTIONSENTRY']._serialized_start=635
  _globals['_MODEL_OPTIONSENTRY']._serialized_end=681
  _globals['_MODEL_MODELRESULT']._serialized_start=1295
  _globals['_MODEL_MODELRESULT']._serialized_end=1403
  _globals['_BIGQUERYSINK']._serialized_start=1406
  _globals['_BIGQUERYSINK']._serialized_end=1641
  _globals['_BIGQUERYSINK_OPTIONSENTRY']._serialized_start=635
  _globals['_BIGQUERYSINK_OPTIONSENTRY']._serialized_end=681
  _globals['_GCSSINK']._serialized_start=1644
  _globals['_GCSSINK']._serialized_end=1890
  _globals['_GCSSINK_OPTIONSENTRY']._serialized_start=635
  _globals['_GCSSINK_OPTIONSENTRY']._serialized_end=681
  _globals['_MAXCOMPUTESINK']._serialized_start=1893
  _globals['_MAXCOMPUTESINK']._serialized_end=2126
  _globals['_MAXCOMPUTESINK_OPTIONSENTRY']._serialized_start=635
  _globals['_MAXCOMPUTESINK_OPTIONSENTRY']._serialized_end=681
# @@protoc_insertion_point(module_scope)



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from pyspark.sql import DataFrame

from merlinpyspark.config import SinkConfig, BigQuerySinkConfig, MaxComputeSinkConfig
import os

from pyspark.sql import SparkSession

def create_sink(spark_session: SparkSession, sink_config: SinkConfig) -> "Sink":
    sink_type = sink_config.sink_type()
    if sink_type == BigQuerySinkConfig.TYPE:
        if not isinstance(sink_config, BigQuerySinkConfig):
            raise ValueError("sink_config is not BigQuerySink")

        return BigQuerySink(sink_config)

    if sink_type == MaxComputeSinkConfig.TYPE:
        if not isinstance(sink_config, MaxComputeSinkConfig):
            raise ValueError("sink_config is not MaxComputeSink")

        return MaxComputeSink(spark_session, sink_config)

    raise ValueError(f"sink type is not implemented: {sink_config.sink_type()}")


class Sink(ABC):
    @abstractmethod
    def save(self, df):
        pass


class BigQuerySink(Sink):
    WRITE_FORMAT = "bigquery"
    OPTION_TABLE = "table"
    OPTION_STAGING_BUCKET = "temporaryGcsBucket"

    def __init__(self, config: BigQuerySinkConfig):
        self._config = config

    def save(self, df: DataFrame):
        df.write.mode(self._config.save_mode()).format(self.WRITE_FORMAT).option(
            self.OPTION_TABLE, self._config.table()
        ).option(self.OPTION_STAGING_BUCKET, self._config.staging_bucket()).options(
            **self._config.options()
        ).save()


class MaxComputeSink(Sink):
    WRITE_FORMAT = "jdbc"
    OPTION_URL = "url"
    OPTION_DRIVER = "driver"
    OPTION_QUERY_TIMEOUT = "queryTimeout"
    OPTION_DB_TABLE = "dbtable"
    OPTION_BATCH_SIZE = "batchSize"

    def __init__(self, spark_session: SparkSession, config: MaxComputeSinkConfig):
        self._config = config
        self._spark = spark_session

    def get_jdbc_url(self):
        url = f"jdbc:odps:{self._config.endpoint()}?project={self._config.project()}&accessId={self.get_access_id()}&accessKey={self.get_access_key()}&interactiveMode={self.get_interactive_mode()}&odpsNamespaceSchema=true&schema={self._config.schema()}&enableLimit=false&alwaysFallback=true&enableOdpsLogger=true"
        if self._config.execute_project() is not None:
            url += f"&executeProject={self._config.execute_project()}"
        return url

    def get_query_timeout(self):
        return self._config.options().get("queryTimeout", "300")

    def get_interactive_mode(self):
        return self._config.options().get("interactiveMode", "true")    

    def get_batch_size(self):
        return self._config.options().get("batchSize", "10000")


    def get_jdbc_driver(self):
        return os.environ.get("ODPS_JDBC_DRIVER", "com.aliyun.odps.jdbc.OdpsDriver")

    def get_access_id(self):
        # NOTE: access id and key will not be part of the Prediction Job proto message
        # these should be passed in via environment variable
        return os.environ.get("ODPS_ACCESS_ID")

    def get_access_key(self):
        # NOTE: access id and key will not be part of the Prediction Job proto message
        # these should be passed in via environment variable
        return os.environ.get("ODPS_SECRET_KEY")
    
    def _get_custom_dialect_class(self):
        # NOTE: this is hardcoded because the gateway function that registers the dialect
        # cannot be passed in at runtime:
        # gw.jvm.org.apache.spark.sql.jdbc.JdbcDialects.registerDialect(
        #    gw.jvm.dev.caraml.spark.odps.CustomDialect()
        # )
        return "dev.caraml.spark.odps.CustomDialect"

    def save(self, df: DataFrame):
        from py4j.java_gateway import java_import
        cfg = self._config

        gw = self._spark.sparkContext._gateway # type: ignore
        java_import(gw.jvm, self._get_custom_dialect_class())
        gw.jvm.org.apache.spark.sql.jdbc.JdbcDialects.registerDialect(
            gw.jvm.dev.caraml.spark.odps.CustomDialect()
        )

        df.write.mode(self._config.save_mode()).format(self.WRITE_FORMAT).option(
            self.OPTION_DRIVER, self.get_jdbc_driver()
        ).option(self.OPTION_URL, self.get_jdbc_url()).option(
            # TODO: use query timeout from config.options() if present
            self.OPTION_QUERY_TIMEOUT,
            self.get_query_timeout(),
        ).option(
            self.OPTION_DB_TABLE, cfg.table()
        ).option(self.OPTION_BATCH_SIZE, self.get_batch_size()).save()



#!/usr/bin/python
# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os

from merlinpyspark.config import load
from merlinpyspark.model import create_model_udf
from merlinpyspark.sink import create_sink
from merlinpyspark.source import create_source
from mlflow import pyfunc
from pyspark import SparkConf, SparkContext
from pyspark.sql import SparkSession

try:
    import pyspark
except ImportError:
    import findspark

    findspark.init()
    import pyspark

DEFAULT_PARALLELISM = 2

parser = argparse.ArgumentParser(description="Run a PySpark job")
parser.add_argument(
    "--job-name",
    type=str,
    required=False,
    dest="job_name",
    help="The name of PySpark job",
    default="merlin-prediction-job",
)
parser.add_argument(
    "--spec-path",
    type=str,
    required=False,
    dest="spec_path",
    help="Path to prediction job yaml file",
)
parser.add_argument(
    "--dry-run-model",
    type=str,
    required=False,
    dest="dry_run_path",
    help="Path to model for dry run",
    default=None,
)
parser.add_argument(
    "--local",
    dest="local",
    action="store_true",
    required=False,
    help="flag to run locally",
    default=False,
)

args, _ = parser.parse_known_args()
print(f"Called with arguments: {args}")


def local_spark_session():
    conf = SparkConf()
    conf.set(
        "spark.jars",
        "https://storage.googleapis.com/hadoop-lib/gcs/gcs-connector"
        "-hadoop2-2.0.1.jar",
    )
    conf.set(
        "spark.jars.packages",
        "com.google.cloud.spark:spark-bigquery-with-dependencies_2.12:0.13.1-beta",
    )

    conf.set(
        "spark.jars",
        "https://github.com/caraml-dev/caraml-store/releases/download/caraml-store-spark/odps/v0.0.1-test/custom-dialect.jar,"
        "https://github.com/aliyun/aliyun-odps-jdbc/releases/download/v3.8.2/odps-jdbc-3.8.2-jar-with-dependencies.jar,"
        "https://storage.googleapis.com/hadoop-lib/gcs/gcs-connector-hadoop2-2.0.1.jar"
    )

    sc = SparkContext(conf=conf)
    sc.setLogLevel("INFO")
    sc._jsc.hadoopConfiguration().set(
        "fs.gs.impl", "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem"
    )
    sc._jsc.hadoopConfiguration().set(
        "fs.AbstractFileSystem.gs.impl",
        "com.google.cloud.hadoop.fs.gcs.GoogleHadoopFS",
    )
    sc._jsc.hadoopConfiguration().set(
        "google.cloud.auth.service.account.enable", "true"
    )

    sa_path = os.environ.get("GOOGLE_APPLICATION_CREDENTIALS")
    if sa_path is None:
        print("You must set GOOGLE_APPLICATION_CREDENTIALS to run locally")

    sc._jsc.hadoopConfiguration().set(
        "google.cloud.auth.service.account.json.keyfile", sa_path
    )

    spark = SparkSession.builder.config(conf=sc.getConf()).getOrCreate()
    return spark


def start(spec_path, spark):
    print(f"loading prediction job spec from: {spec_path}")
    job_spec = load(spec_path)

    # target parallelism is determined by number of executor instances * number of core per executor
    # for local case spark.executor.instances and spark.executor.cores are None and will fallback to DEFAULT_PARALLELISM
    instances = spark.sparkContext.getConf().get("spark.executor.instances")
    cores = spark.sparkContext.getConf().get("spark.executor.cores")
    target_parallelism = DEFAULT_PARALLELISM  # to handle local test
    if instances is not None and cores is not None:
        target_parallelism = int(instances) * int(cores)

    print(f"target parallelism: {target_parallelism}")

    data_source = create_source(spark, job_spec.source())
    df = data_source.load()
    features = list(data_source.features())
    if features is None:
        features = df.columns

    current_partition = df.rdd.getNumPartitions()
    if target_parallelism > current_partition:
        # Repartition the dataframe to have same number of partition as target_parallelism for better executor utilization
        print(
            f"repartition data from {df.rdd.getNumPartitions()} to {target_parallelism}"
        )
        df = df.repartition(target_parallelism)
    elif current_partition > target_parallelism:
        target_partition = (
            current_partition
            - (current_partition % target_parallelism)
            + target_parallelism
        )
        print(f"repartition data from {current_partition} to {target_partition}")
        df = df.repartition(target_partition)

    model_udf = create_model_udf(spark, job_spec.model(), features)
    data_sink = create_sink(spark, job_spec.sink())

    df = df.withColumn(job_spec.sink().result_column(), model_udf(*features))

    data_sink.save(df)

    print(f"The prediction job completed successfully!")


def main():
    if args.dry_run_path is not None:
        pyfunc.load_model(args.dry_run_path)
        print("Dry run success")
        exit(0)

    if args.local:
        spark = local_spark_session()
    else:
        spark = SparkSession.builder.appName(args.job_name).getOrCreate()
        spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")

    print(f"Spark configuration: {spark.sparkContext.getConf().getAll()}")

    start(args.spec_path, spark)


if __name__ == "__main__":
    main()



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from pyspark.sql.types import ArrayType, DoubleType, FloatType, IntegerType, \
    LongType, StringType

from merlinpyspark.config import BigQuerySinkConfig, BigQuerySourceConfig, load
from test.util.test_utils import write_config

sample_1_yaml = """
kind: PredictionJob
version: v1
name: sample-1
bigquerySource:
  table: 'bigquery-public-data:samples.shakespeare'
  features:
    - word
    - word_count
  options:
    filter: "word_count > 100"
model:
  type: PYFUNC_V2
  uri: "gs://bucket-name/mlflow/6/2c3703fbbf9f4866b26e4cf91641f02c/artifacts/model"
  result:
    type: ARRAY
    item_type: DOUBLE
bigquerySink:
  table: 'project.dataset.table'
  result_column: 'prediction'
  save_mode: OVERWRITE
  options:
    project: "project"
    temporaryGcsBucket: "bucket-name"
"""

sample_2_yaml = """
kind: PredictionJob
version: v1
name: sample-2
bigquerySource:
  table: 'bigquery-public-data:samples.shakespeare'
model:
  type: PYFUNC_V2
  uri: "gs://bucket-name/mlflow/6/2c3703fbbf9f4866b26e4cf91641f02c/artifacts/model"
bigquerySink:
  table: 'project.dataset.table'
"""

test_type_yaml = """
kind: PredictionJob
version: v1
name: test-type
bigquerySource:
  table: 'bigquery-public-data:samples.shakespeare'
model:
  type: PYFUNC_V2
  uri: "gs://bucket-name/mlflow/6/2c3703fbbf9f4866b26e4cf91641f02c/artifacts/model"
  result:
    type: {}
    item_type: {}
bigquerySink:
  table: 'project.dataset.table'
"""


@pytest.fixture
def sample_1_path():
    return write_config("test-config/sample_1.yaml", sample_1_yaml)


@pytest.fixture
def sample_2_path():
    return write_config("test-config/sample_2.yaml", sample_2_yaml)


def test_load_1(sample_1_path):
    cfg = load(sample_1_path)
    assert cfg is not None
    assert cfg.model() is not None
    assert cfg.source() is not None
    assert cfg.sink() is not None

    src_cfg = cfg.source()
    assert src_cfg.source_type() == "bigquery"
    assert src_cfg.options()["filter"] == "word_count > 100"
    assert len(src_cfg.features()) == 2
    assert src_cfg.features()[0] == "word"
    assert src_cfg.features()[1] == "word_count"
    assert isinstance(src_cfg, BigQuerySourceConfig)
    assert src_cfg.table() == "bigquery-public-data:samples.shakespeare"

    model_cfg = cfg.model()
    assert model_cfg.model_type() == "PYFUNC_V2"
    assert model_cfg.result_type() == ArrayType(DoubleType())
    assert model_cfg.model_uri() == "gs://bucket-name/mlflow/6/2c3703fbbf9f4866b26e4cf91641f02c/artifacts/model"

    sink_cfg = cfg.sink()
    assert sink_cfg.sink_type() == "bigquery"
    assert sink_cfg.save_mode() == "overwrite"
    assert sink_cfg.result_column() == "prediction"
    assert sink_cfg.options()["temporaryGcsBucket"] == "bucket-name"
    assert isinstance(sink_cfg, BigQuerySinkConfig)
    assert sink_cfg.table() == "project.dataset.table"


def test_load_2(sample_2_path):
    cfg = load(sample_2_path)
    assert cfg.model() is not None
    assert cfg.source() is not None
    assert cfg.sink() is not None

    src_cfg = cfg.source()
    assert src_cfg.source_type() == "bigquery"
    assert len(src_cfg.options()) == 0
    assert len(src_cfg.features()) == 0
    assert isinstance(src_cfg, BigQuerySourceConfig)
    assert src_cfg.table() == "bigquery-public-data:samples.shakespeare"

    model_cfg = cfg.model()
    assert model_cfg.model_type() == "PYFUNC_V2"
    assert model_cfg.result_type() == DoubleType()
    assert model_cfg.model_uri() == "gs://bucket-name/mlflow/6/2c3703fbbf9f4866b26e4cf91641f02c/artifacts/model"

    sink_cfg = cfg.sink()
    assert sink_cfg.sink_type() == "bigquery"
    assert sink_cfg.save_mode() == "errorifexists"
    assert sink_cfg.result_column() == "prediction"
    assert len(sink_cfg.options()) == 0
    assert isinstance(sink_cfg, BigQuerySinkConfig)
    assert sink_cfg.table() == "project.dataset.table"


@pytest.mark.parametrize(
    "return_type,item_type,expected",
    [
        ("FLOAT", "", FloatType()),
        ("DOUBLE", "", DoubleType()),
        ("STRING", "", StringType()),
        ("INTEGER", "", IntegerType()),
        ("LONG", "", LongType()),
        ("ARRAY", "FLOAT", ArrayType(FloatType())),
        ("ARRAY", "DOUBLE", ArrayType(DoubleType())),
        ("ARRAY", "STRING", ArrayType(StringType())),
        ("ARRAY", "INTEGER", ArrayType(IntegerType())),
        ("ARRAY", "LONG", ArrayType(LongType())),
    ]
)
def test_model_result_conversion(return_type, item_type, expected):
    cfg_file = write_config("test-config/test_type.yaml",
                            test_type_yaml.format(
                                return_type,
                                item_type))

    cfg = load(cfg_file)
    assert cfg.model().result_type() == expected





"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 The Merlin Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _ResultType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ResultTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ResultType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DOUBLE: _ResultType.ValueType  # 0
    FLOAT: _ResultType.ValueType  # 1
    INTEGER: _ResultType.ValueType  # 2
    LONG: _ResultType.ValueType  # 3
    STRING: _ResultType.ValueType  # 4
    ARRAY: _ResultType.ValueType  # 10

class ResultType(_ResultType, metaclass=_ResultTypeEnumTypeWrapper): ...

DOUBLE: ResultType.ValueType  # 0
FLOAT: ResultType.ValueType  # 1
INTEGER: ResultType.ValueType  # 2
LONG: ResultType.ValueType  # 3
STRING: ResultType.ValueType  # 4
ARRAY: ResultType.ValueType  # 10
global___ResultType = ResultType

class _ModelType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ModelTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ModelType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    INVALID_MODEL_TYPE: _ModelType.ValueType  # 0
    XGBOOST: _ModelType.ValueType  # 1
    TENSORFLOW: _ModelType.ValueType  # 2
    SKLEARN: _ModelType.ValueType  # 3
    PYTORCH: _ModelType.ValueType  # 4
    ONNX: _ModelType.ValueType  # 5
    PYFUNC: _ModelType.ValueType  # 6
    PYFUNC_V2: _ModelType.ValueType  # 7

class ModelType(_ModelType, metaclass=_ModelTypeEnumTypeWrapper): ...

INVALID_MODEL_TYPE: ModelType.ValueType  # 0
XGBOOST: ModelType.ValueType  # 1
TENSORFLOW: ModelType.ValueType  # 2
SKLEARN: ModelType.ValueType  # 3
PYTORCH: ModelType.ValueType  # 4
ONNX: ModelType.ValueType  # 5
PYFUNC: ModelType.ValueType  # 6
PYFUNC_V2: ModelType.ValueType  # 7
global___ModelType = ModelType

class _FileFormat:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _FileFormatEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_FileFormat.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    INVALID_FILE_FORMAT: _FileFormat.ValueType  # 0
    CSV: _FileFormat.ValueType  # 1
    PARQUET: _FileFormat.ValueType  # 2
    AVRO: _FileFormat.ValueType  # 3
    JSON: _FileFormat.ValueType  # 4

class FileFormat(_FileFormat, metaclass=_FileFormatEnumTypeWrapper): ...

INVALID_FILE_FORMAT: FileFormat.ValueType  # 0
CSV: FileFormat.ValueType  # 1
PARQUET: FileFormat.ValueType  # 2
AVRO: FileFormat.ValueType  # 3
JSON: FileFormat.ValueType  # 4
global___FileFormat = FileFormat

class _SaveMode:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _SaveModeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_SaveMode.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    ERRORIFEXISTS: _SaveMode.ValueType  # 0
    OVERWRITE: _SaveMode.ValueType  # 1
    APPEND: _SaveMode.ValueType  # 2
    IGNORE: _SaveMode.ValueType  # 3
    ERROR: _SaveMode.ValueType  # 4

class SaveMode(_SaveMode, metaclass=_SaveModeEnumTypeWrapper): ...

ERRORIFEXISTS: SaveMode.ValueType  # 0
OVERWRITE: SaveMode.ValueType  # 1
APPEND: SaveMode.ValueType  # 2
IGNORE: SaveMode.ValueType  # 3
ERROR: SaveMode.ValueType  # 4
global___SaveMode = SaveMode

@typing.final
class PredictionJob(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VERSION_FIELD_NUMBER: builtins.int
    KIND_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    BIGQUERY_SOURCE_FIELD_NUMBER: builtins.int
    GCS_SOURCE_FIELD_NUMBER: builtins.int
    MAXCOMPUTE_SOURCE_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    BIGQUERY_SINK_FIELD_NUMBER: builtins.int
    GCS_SINK_FIELD_NUMBER: builtins.int
    MAXCOMPUTE_SINK_FIELD_NUMBER: builtins.int
    version: builtins.str
    kind: builtins.str
    name: builtins.str
    @property
    def bigquery_source(self) -> global___BigQuerySource: ...
    @property
    def gcs_source(self) -> global___GcsSource: ...
    @property
    def maxcompute_source(self) -> global___MaxComputeSource: ...
    @property
    def model(self) -> global___Model: ...
    @property
    def bigquery_sink(self) -> global___BigQuerySink: ...
    @property
    def gcs_sink(self) -> global___GcsSink: ...
    @property
    def maxcompute_sink(self) -> global___MaxComputeSink: ...
    def __init__(
        self,
        *,
        version: builtins.str = ...,
        kind: builtins.str = ...,
        name: builtins.str = ...,
        bigquery_source: global___BigQuerySource | None = ...,
        gcs_source: global___GcsSource | None = ...,
        maxcompute_source: global___MaxComputeSource | None = ...,
        model: global___Model | None = ...,
        bigquery_sink: global___BigQuerySink | None = ...,
        gcs_sink: global___GcsSink | None = ...,
        maxcompute_sink: global___MaxComputeSink | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bigquery_sink", b"bigquery_sink", "bigquery_source", b"bigquery_source", "gcs_sink", b"gcs_sink", "gcs_source", b"gcs_source", "maxcompute_sink", b"maxcompute_sink", "maxcompute_source", b"maxcompute_source", "model", b"model", "sink", b"sink", "source", b"source"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bigquery_sink", b"bigquery_sink", "bigquery_source", b"bigquery_source", "gcs_sink", b"gcs_sink", "gcs_source", b"gcs_source", "kind", b"kind", "maxcompute_sink", b"maxcompute_sink", "maxcompute_source", b"maxcompute_source", "model", b"model", "name", b"name", "sink", b"sink", "source", b"source", "version", b"version"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["sink", b"sink"]) -> typing.Literal["bigquery_sink", "gcs_sink", "maxcompute_sink"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["source", b"source"]) -> typing.Literal["bigquery_source", "gcs_source", "maxcompute_source"] | None: ...

global___PredictionJob = PredictionJob

@typing.final
class BigQuerySource(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    TABLE_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    table: builtins.str
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        table: builtins.str = ...,
        features: collections.abc.Iterable[builtins.str] | None = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["features", b"features", "options", b"options", "table", b"table"]) -> None: ...

global___BigQuerySource = BigQuerySource

@typing.final
class GcsSource(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    FORMAT_FIELD_NUMBER: builtins.int
    URI_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    format: global___FileFormat.ValueType
    uri: builtins.str
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        format: global___FileFormat.ValueType = ...,
        uri: builtins.str = ...,
        features: collections.abc.Iterable[builtins.str] | None = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["features", b"features", "format", b"format", "options", b"options", "uri", b"uri"]) -> None: ...

global___GcsSource = GcsSource

@typing.final
class MaxComputeSource(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    TABLE_FIELD_NUMBER: builtins.int
    ENDPOINT_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    table: builtins.str
    endpoint: builtins.str
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        table: builtins.str = ...,
        endpoint: builtins.str = ...,
        features: collections.abc.Iterable[builtins.str] | None = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["endpoint", b"endpoint", "features", b"features", "options", b"options", "table", b"table"]) -> None: ...

global___MaxComputeSource = MaxComputeSource

@typing.final
class Model(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing.final
    class ModelResult(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TYPE_FIELD_NUMBER: builtins.int
        ITEM_TYPE_FIELD_NUMBER: builtins.int
        type: global___ResultType.ValueType
        item_type: global___ResultType.ValueType
        """only if type is array"""
        def __init__(
            self,
            *,
            type: global___ResultType.ValueType = ...,
            item_type: global___ResultType.ValueType = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["item_type", b"item_type", "type", b"type"]) -> None: ...

    TYPE_FIELD_NUMBER: builtins.int
    URI_FIELD_NUMBER: builtins.int
    RESULT_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    type: global___ModelType.ValueType
    uri: builtins.str
    @property
    def result(self) -> global___Model.ModelResult: ...
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        type: global___ModelType.ValueType = ...,
        uri: builtins.str = ...,
        result: global___Model.ModelResult | None = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["result", b"result"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["options", b"options", "result", b"result", "type", b"type", "uri", b"uri"]) -> None: ...

global___Model = Model

@typing.final
class BigQuerySink(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    TABLE_FIELD_NUMBER: builtins.int
    STAGING_BUCKET_FIELD_NUMBER: builtins.int
    RESULT_COLUMN_FIELD_NUMBER: builtins.int
    SAVE_MODE_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    table: builtins.str
    staging_bucket: builtins.str
    result_column: builtins.str
    save_mode: global___SaveMode.ValueType
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        table: builtins.str = ...,
        staging_bucket: builtins.str = ...,
        result_column: builtins.str = ...,
        save_mode: global___SaveMode.ValueType = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["options", b"options", "result_column", b"result_column", "save_mode", b"save_mode", "staging_bucket", b"staging_bucket", "table", b"table"]) -> None: ...

global___BigQuerySink = BigQuerySink

@typing.final
class GcsSink(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    FORMAT_FIELD_NUMBER: builtins.int
    URI_FIELD_NUMBER: builtins.int
    RESULT_COLUMN_FIELD_NUMBER: builtins.int
    SAVE_MODE_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    format: global___FileFormat.ValueType
    uri: builtins.str
    result_column: builtins.str
    save_mode: global___SaveMode.ValueType
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        format: global___FileFormat.ValueType = ...,
        uri: builtins.str = ...,
        result_column: builtins.str = ...,
        save_mode: global___SaveMode.ValueType = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["format", b"format", "options", b"options", "result_column", b"result_column", "save_mode", b"save_mode", "uri", b"uri"]) -> None: ...

global___GcsSink = GcsSink

@typing.final
class MaxComputeSink(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OptionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    TABLE_FIELD_NUMBER: builtins.int
    ENDPOINT_FIELD_NUMBER: builtins.int
    RESULT_COLUMN_FIELD_NUMBER: builtins.int
    SAVE_MODE_FIELD_NUMBER: builtins.int
    OPTIONS_FIELD_NUMBER: builtins.int
    table: builtins.str
    endpoint: builtins.str
    result_column: builtins.str
    save_mode: global___SaveMode.ValueType
    @property
    def options(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        table: builtins.str = ...,
        endpoint: builtins.str = ...,
        result_column: builtins.str = ...,
        save_mode: global___SaveMode.ValueType = ...,
        options: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["endpoint", b"endpoint", "options", b"options", "result_column", b"result_column", "save_mode", b"save_mode", "table", b"table"]) -> None: ...

global___MaxComputeSink = MaxComputeSink



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Union

import joblib
import mlflow
import numpy as np
import pandas as pd
import pytest
from merlin.model import PyFuncV2Model
from merlinpyspark.config import ModelConfig
from merlinpyspark.model import create_model_udf
from merlinpyspark.spec.prediction_job_pb2 import Model, ModelType, ResultType
from mlflow.pyfunc import log_model
from pandas.testing import assert_frame_equal
from sklearn.datasets import load_iris


class IrisModel(PyFuncV2Model):
    def initialize(self, artifacts: dict):
        self._model = joblib.load(artifacts["model_path"])

    def infer(
        self, model_input: pd.DataFrame
    ) -> Union[np.ndarray, pd.Series, pd.DataFrame]:
        return self._model.predict(model_input)


@pytest.mark.ci
def test_pyfuncv2_model(spark_session):
    log_model(
        "model",
        python_model=IrisModel(),
        artifacts={"model_path": "test-model/artifacts/model.joblib"},
        code_path=["merlinpyspark", "test"],
        conda_env="test-model/conda.yaml",
    )
    model_path = os.path.join(mlflow.get_artifact_uri(), "model")

    iris = load_iris()
    columns = ["sepal_length", "sepal_width", "petal_length", "petal_width"]
    pdf = pd.DataFrame(iris.data, columns=columns)
    df = spark_session.createDataFrame(pdf)

    model_cfg_proto = Model(
        type=ModelType.PYFUNC_V2,
        uri=model_path,
        result=Model.ModelResult(type=ResultType.DOUBLE),
    )

    model_udf = create_model_udf(spark_session, ModelConfig(model_cfg_proto), columns)
    iris_model = mlflow.pyfunc.load_model(model_path)

    df = df.withColumn(
        "prediction",
        model_udf("sepal_length", "sepal_width", "petal_length", "petal_width"),
    )

    result_pandas = df.toPandas()
    r = result_pandas.drop(columns=["prediction"])

    assert_frame_equal(r, pdf)
    assert np.array_equal(result_pandas["prediction"], iris_model.predict(pdf))
    assert len(df.columns) == len(columns) + 1



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Iterable, List

from pyspark.sql import DataFrame, SparkSession
import os

from merlinpyspark.config import (
    BigQuerySourceConfig,
    SourceConfig,
    MaxComputeSourceConfig,
)


def create_source(spark_session: SparkSession, source_config: SourceConfig) -> "Source":
    source_type = source_config.source_type()

    if source_type == BigQuerySourceConfig.TYPE:
        if not isinstance(source_config, BigQuerySourceConfig):
            raise ValueError("source_config is not of BigQuerySourceConfig instance")
        return BigQuerySource(spark_session, source_config)

    if source_type == MaxComputeSourceConfig.TYPE:
        if not isinstance(source_config, MaxComputeSourceConfig):
            raise ValueError("source_config is not of MaxComputeSourceConfig instance")
        return MaxComputeSource(spark_session, source_config)

    raise NotImplementedError(f"{source_type} source type is not implemented")


class Source(ABC):
    @abstractmethod
    def load(self) -> DataFrame:
        """
        Loads a data stream from a data source and returns it as a :class`DataFrame`.
        :return: Spark DataFrame
        """
        pass

    @abstractmethod
    def features(self) -> Iterable[str]:
        """
        List of columns used for prediction
        :return: list of string
        """
        pass


class BigQuerySource(Source):
    READ_FORMAT = "bigquery"
    OPTION_TABLE = "table"
    OPTION_PARALLELISM = "maxParallelism"

    def __init__(
        self, spark_session: SparkSession, bq_source_config: BigQuerySourceConfig
    ):
        self._spark = spark_session
        self._config = bq_source_config

        if bq_source_config.table() is None:
            raise ValueError("table field is empty")

    def load(self) -> DataFrame:
        cfg = self._config
        features = self.features()

        reader = self._spark.read.format(self.READ_FORMAT).option(
            self.OPTION_TABLE, cfg.table()
        )

        if cfg.options() is not None:
            reader.options(**cfg.options())

        df = reader.load()
        if features is not None:
            return df.select(*features)

        return df

    def features(self) -> Iterable[str]:
        return self._config.features()


class MaxComputeSource(Source):
    READ_FORMAT = "jdbc"
    OPTION_URL = "url"
    OPTION_DRIVER = "driver"
    OPTION_QUERY_TIMEOUT = "queryTimeout"
    OPTION_DB_TABLE = "dbtable"
    OPTION_FETCH_SIZE = "fetchSize"
    OPTION_SESSION_INIT_STATEMENT = "sessionInitStatement"

    def __init__(
        self,
        spark_session: SparkSession,
        maxcompute_source_config: MaxComputeSourceConfig,
    ):
        self._spark = spark_session
        self._config = maxcompute_source_config
        self.auto_select_limit = os.environ.get(
            "CARAML_SPARK_MAXCOMPUTE_AUTO_SELECT_LIMIT", "1000000000"
        )

        if maxcompute_source_config.table() is None:
            raise ValueError("table field is empty")

        if maxcompute_source_config.project() is None:
            raise ValueError("project field is empty")

        if maxcompute_source_config.schema() is None:
            raise ValueError("schema field is empty")

    def load(self) -> DataFrame:
        from py4j.java_gateway import java_import

        gw = self._spark.sparkContext._gateway  # type: ignore
        java_import(gw.jvm, self._get_custom_dialect_class())
        gw.jvm.org.apache.spark.sql.jdbc.JdbcDialects.registerDialect(
            gw.jvm.dev.caraml.spark.odps.CustomDialect()
        )
        cfg = self._config
        reader = (
            self._spark.read.format(self.READ_FORMAT)
            .option(self.OPTION_DRIVER, self.get_jdbc_driver())
            .option(self.OPTION_URL, self.get_jdbc_url())
            .option(self.OPTION_QUERY_TIMEOUT, self.get_query_timeout())
            .option(self.OPTION_DB_TABLE, cfg.table())
            .option(self.OPTION_FETCH_SIZE, self.get_fetch_size())
            .option(
                self.OPTION_SESSION_INIT_STATEMENT, "SET odps.sql.allow.fullscan=true;"
            )  # to allow fullscan table on partitioned table
        )
        if cfg.options() is not None:
            reader.options(**cfg.options())
        df = reader.load()
        if self.features() is not None:
            return df.select(*self.features())
        return df

    def features(self) -> Iterable[str]:
        return self._config.features()

    def get_jdbc_url(self):
        url = f"jdbc:odps:{self._config.endpoint()}?project={self._config.project()}&accessId={self.get_access_id()}&accessKey={self.get_access_key()}&interactiveMode={self.get_interactive_mode()}&odpsNamespaceSchema=true&schema={self._config.schema()}&enableLimit=false&autoSelectLimit={self.auto_select_limit}&alwaysFallback=true&enableOdpsLogger=true"
        if self._config.execute_project() is not None:
            url += f"&executeProject={self._config.execute_project()}"
        return url

    def get_query_timeout(self):
        return self._config.options().get("queryTimeout", "300")

    def get_interactive_mode(self):
        return self._config.options().get("interactiveMode", "true")

    def get_fetch_size(self):
        return self._config.options().get("fetchSize", "0")

    def get_jdbc_driver(self):
        return os.environ.get("ODPS_JDBC_DRIVER", "com.aliyun.odps.jdbc.OdpsDriver")

    def get_access_id(self):
        # NOTE: access id and key will not be part of the PredictionConfig
        # since these are mounted from a configmap
        # these should be passed in via environment variable
        return os.environ.get("ODPS_ACCESS_ID")

    def get_access_key(self):
        # NOTE: access id and key will not be part of the PredictionConfig
        # since these are mounted from a configmap
        # these should be passed in via environment variable
        return os.environ.get("ODPS_SECRET_KEY")

    def _get_custom_dialect_class(self):
        # NOTE: this is hardcoded because of how it should be imported in the spark context
        return "dev.caraml.spark.odps.CustomDialect"



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import pytest

from merlinpyspark.config import BigQuerySinkConfig
from merlinpyspark.sink import create_sink
from merlinpyspark.spec.prediction_job_pb2 import BigQuerySink


@pytest.mark.ci
def test_bq_sink(spark_session, bq):
    bq_sink_proto = BigQuerySink(table='project.dataset.table',
                                 result_column="prediction",
                                 staging_bucket="bucket-name",
                                 options={"project": "project"})

    sink_cfg = BigQuerySinkConfig(bq_sink_proto)
    table_id = sink_cfg.table()

    bq.delete_table(table=table_id, not_found_ok=True)

    test_data = {
        "word": ["lorem", "ipsum"],
        "word_count": [1, 2]
    }
    pdf = pd.DataFrame(test_data)
    df = spark_session.createDataFrame(pdf)

    bq_sink = create_sink(sink_cfg)
    bq_sink.save(df)

    t = bq.get_table(table_id)
    assert len(t.schema) == 2
    assert t.schema[0].name == 'word'
    assert t.schema[0].field_type == 'STRING'
    assert t.schema[1].name == 'word_count'
    assert t.schema[1].field_type == 'INTEGER'
